/*
 * lftp and utils
 *
 * Copyright (c) 1996-1997 by Alexander V. Lukyanov (lav@yars.free.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <config.h>

#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <ctype.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "misc.h"
#include "CmdExec.h"
#include "GetJob.h"
#include "PutJob.h"
#include "CatJob.h"
#include "LsJob.h"
#include "LsCache.h"
#include "mgetJob.h"
#include "mputJob.h"
#include "mkdirJob.h"
#include "rmJob.h"
#include "SysCmdJob.h"
#include "SiteJob.h"
#include "MirrorJob.h"
#include "mvJob.h"

#include "alias.h"
#include "netrc.h"
#include "url.h"
#include "lftp.h"
#include "vars.h"
#include "GetPass.h"

const struct CmdExec::cmd_rec CmdExec::cmd_table[]=
{
   {"!",       do_shell,   "!<shell_command>",
	 "Launch shell or shell command\n"},
   {"(",       do_subsh,   "(commands)",
	 "Group commands together to be executed as one command\n"
	 "You can launch such a group in background\n"},
   {"?",       do_help,	   0,"help"},
   {"alias",   do_alias,   "alias [<name> [<value>]]",
	 "Define or undefine alias <name>. If <value> omitted,\n"
	 "the alias is undefined, else is takes the value <value>\n"},
   {"anon",    do_anon,    "anon",
	 "anon - login anonymously (by default)\n"},
   {"bye",     do_exit,	   0,"exit"},
   {"cache",   do_cache,   "cache [SUBCMD]",
	 "cache command controls local memory cache\n\n"
	 "The following subcommands are recognized:\n"
	 "  stat       - print cache status (default)\n"
	 "  on|off     - turn on/off caching\n"
	 "  flush      - flush cache\n"
	 "  size <lim> - set memory limit, -1 means unlimited\n"},
   {"cat",     do_cat,     "cat <files>",
	 "cat <files> - output remote files to stdin\n"},
   {"cd",      do_cd,      "cd <rdir>",
	 "Change current remote directory\n"},
   {"connect", do_open,	   0,"open"},
   {"debug",   do_debug,   "debug <level>|off"},
   {"exit",    do_exit,    "exit [<code>]",
	 "exit - exit from lftp or move to background if jobs are active\n\n"
	 "If no jobs active, the code is passed to operating system as lftp\n"
	 "termination status. If omitted, exit code of last command is used.\n"},
   {"get",     do_get,  "get [-c] <rfile> [-o <lfile>]",
	 "Retrieve remote file <rfile> and store it to local file <lfile>.\n"
	 "If -o omitted, the file is stored to local file named as basename\n"
	 "of <rfile>. If option [-c] is specified, try to reget specified files.\n"},
   {"help",    do_help,    "help [<cmd>]",
	 "Print help for command <cmd>, or list of available commands\n"},
   {"jobs",    do_jobs,    "jobs [-v]",
	 "List running jobs. -v means verbose, several -v can be specified.\n"},
   {"kill",    do_kill,    "kill all|<job_no>",
	 "Delete specified job with <job_no> or all jobs\n"},
   {"lcd",     do_lcd,     "lcd <ldir>",
	 "Change current local directory <ldir>\n"},
   {"lftp",    do_open,	   0,"open"},
   {"login",   do_user,	   0,"user"},
   {"ls",      do_ls,	   "ls <params>",
	 "List remote files. You can redirect output of this command to file\n"
	 "to file or via pipe to external command.\n"},
   {"mget",    do_mget,    "mget [-c] [-d] <files>",
	 "Gets selected files with expanded wildcards\n"
	 " -c  continue, reget\n"
	 " -d  create directories the same as in file names and get the\n"
	 "     files into them instead of current directory\n"},
   {"mirror",  do_mirror,  "mirror [OPTS] [remote [local]]",
	 "\nMirror specified remote directory to local directory\n\n"
	 " -e, --delete           delete files not present at remote site\n"
	 " -s, --allow-suid       set suid/sgid bits according to remote site\n"
	 " -n, --only-newer       download only newer files\n"
	 " -r, --no-recursion     don't go to subdirectories\n"
	 " -p, --no-perms         don't set file permissions\n"
	 " -i RX, --include RX    include matching files (only one allowed)\n"
	 " -x RX, --exclude RX    exclude matching files (only one allowed)\n"
	 " -t Nx, --time-prec Nx  set time precision to N seconds (x=s)\n"
	 "                        minutes (x=m) hours (x=h) or days (x=d)\n"},
	 
   {"mkdir",   do_mkdir,    "mkdir <dirs>",
	 "Make remote directories\n"},
   {"more",    do_cat,     "more <files>",
	 "Same as `cat <files> | more'. if PAGER is set, it is used as filter\n"},
   {"mput",    do_mput,	   "mput [-c] [-d] <files>",
	 "Upload files matching patterns. -c means reput, -d means to create\n"
	 "remote directories and use the same remote name as local one. By default\n"
	 "It uses basename of local name as remote one.\n"},
   {"mv",      do_mv,	   "mv <file1> <file2>",
	 "Rename <file1> to <file2>\n"},
   {"open",    do_open,    "open <host>",
	 "open - select an ftp server\n"
	 "Usage: open [-e cmd] [-u user[,pass]] [-p port] <host|url>\n"},
   {"put",     do_put,  "put [-c] <lfile> [-o <rfile>]",
	 "Upload <lfile> with remote name <rfile>. If -o omitted, basename\n"
	 "of <lfile> is used as remote name. -c means reput, which requires\n"
	 "you to have permission to overwrite remote files in current directory.\n"
	 "Some of ftp servers (wu-ftpd2.4-beta11, ncftpd-?.?) can't be used\n"
	 "to restart upload because of bugs. In that case use `set norest y'.\n"},
   {"pwd",     do_pwd,     "pwd",
	 "Print current remote directory\n"},
   {"quit",    do_exit,	   0,"exit"},
   {"reget",   do_get,  "reget <rfile> [-o <lfile>]",
	 "Same as `get -c'\n"},
   {"rels",    do_ls,	   "rels [<args>]",
	 "Same as `ls', but don't look in cache\n"},
   {"reput",   do_put,  "reput <lfile> [-o <rfile>]",
	 "Same as `put -c'\n"},
   {"rm",      do_rm,	   "rm <files>",
	 "Remove remote files\n"},
   {"rmdir",   do_rm,	   "rmdir <dirs>",
	 "Remove remote directories\n"},
   {"set",     do_set,     "set [<var> <val>]"},
   {"site",    do_site,	   "site <site_cmd>",
	 "Execute site command <site_cmd> and output the result\n"
	 "You can redirect its output\n"},
   {"source",  do_source,  "source <file>",
	 "Execute commands recorded in file <file>\n"},
   {"user",    do_user,    "user <user> [<pass>]",
	 "Use specified info for remote login\n"},
   {"wait",    do_wait,	   "wait <jobno>",
	 "Wait for spacified job to terminate.\n"},
   {"zcat",    do_cat,     "zcat <files>",
	 "Same as cat, but filter each file trough zcat\n"},
   {"zmore",   do_cat,     "zmore <files>",
	 "Same as more, but filter each file trough zcat\n"},

   {NULL,NULL}
};

CMD(lcd)
{
   if(args->count()<2)
   {
      fprintf(stderr,"Usage: %s local-dir\n",args->getarg(0));
      return 0;
   }
   char *cd_to=expand_home_relative(args->getarg(1));
   RestoreCWD();
   int res=chdir(cd_to);
   if(res==-1)
   {
      perror(cd_to);
      exit_code=1;
      return 0;
   }
   SaveCWD();
   if(interactive)
      printf("lcd ok, local cwd=%s\n",cwd);
   exit_code=0;
   return 0;
}

CMD(ls)
{
   if(args->count()==1)
      args->Append(var_ls);
   LsJob *j=new LsJob(Clone(),output,args->Combine(1));
   output=0;
   if(!strncmp(args->a0(),"re",2))
      j->NoCache();
   return j;
}

CMD(cat)
{
   if(args->count()==1)
   {
      fprintf(stderr,"Usage: %s files...\n",args->a0());
      return 0;
   }
   Job *j=new CatJob(Clone(),output,args);
   output=0;
   args=0;
   return j;
}

char *basename_ptr(char *s)
{
   char *s1=strrchr(s,'/');
   return s1?s1+1:s;
}

char *expand_home_relative(char *s)
{
   if(s[0]=='~')
   {
      char *home=0;
      char *sl;
      static char *ret_path=0;
      
      if(s[1]==0 || s[1]=='/')
      {
	 home=getenv("HOME");
      }
      else
      {
	 char name[9];
	 strncpy(name,s+1,8);
	 name[8]=0;
	 sl=strchr(name,'/');
	 if(sl)
	    *sl=0;
	    
	 struct passwd *pw=getpwnam(name);
	 if(pw)
	    home=pw->pw_dir;
      }
      if(home==0)
	 return s;
      
      sl=strchr(s,'/');
      if(sl)
      {
	 ret_path=(char*)xrealloc(ret_path,strlen(sl)+strlen(home)+1);
	 strcpy(ret_path,home);
	 strcat(ret_path,sl);
	 return ret_path;
      }
      return home;
   }
   return s;
}

CMD(get)
{
   int opt;
   bool cont=false;
   char *opts="+c";
   char *op=args->a0();
   ArgV	 *get_args=new ArgV(op);
   
   args->rewind();
   if(!strncmp(op,"re",2))
   {
      cont=true;
      opts="+";
   }
   while((opt=args->getopt(opts))!=EOF)
   {
      switch(opt)
      {
      case('c'):
	 cont=true;
	 break;
      case('?'):
      err:
	 fprintf(stderr,"Try `help %s' for more information\n",op);
	 return 0;
      }
   }
   args->back();
   char *a=args->getnext();
   char *a1;
   if(a==0)
   {
      fprintf(stderr,"File name missed. ");
      goto err;
   }
   while(a)
   {
      get_args->Append(a);
      a1=basename_ptr(a);
      a=args->getnext();
      if(a && !strcmp(a,"-o"))
      {
	 a=args->getnext();
	 if(a)
	 {
	    a=expand_home_relative(a);
	    struct stat st;
	    int res=stat(a,&st);
	    if(res!=-1 && S_ISDIR(st.st_mode))
	    {
	       char *comb=(char*)alloca(strlen(a)+strlen(a1)+2);
	       sprintf(comb,"%s/%s",a,a1);
	       get_args->Append(comb);
	    }
	    else
	    {
	       get_args->Append(a);
	    }
	 }
	 else
	    get_args->Append(a1);
	 a=args->getnext();
      }
      else
	 get_args->Append(a1);
   }

   GetJob *j=new GetJob(Clone(),get_args,cont);
   return j;
}

CMD(put)
{
   int opt;
   bool cont=false;
   char *opts="+c";
   char *op=args->a0();
   ArgV	 *get_args=new ArgV(op);
   
   args->rewind();
   if(!strncmp(op,"re",2))
   {
      cont=true;
      opts="+";
   }
   while((opt=args->getopt(opts))!=EOF)
   {
      switch(opt)
      {
      case('c'):
	 cont=true;
	 break;
      case('?'):
      err:
	 fprintf(stderr,"Try `help %s' for more information\n",op);
	 return 0;
      }
   }
   args->back();
   char *a=args->getnext();
   char *a1;
   if(a==0)
   {
      fprintf(stderr,"File name missed. ");
      goto err;
   }
   while(a)
   {
      a=expand_home_relative(a);
      get_args->Append(a);
      a1=basename_ptr(a);
      a=args->getnext();
      if(a && !strcmp(a,"-o"))
      {
	 a=args->getnext();
	 if(a)
	    get_args->Append(a);
	 else
	    get_args->Append(a1);
	 a=args->getnext();
      }
      else
	 get_args->Append(a1);
   }

   PutJob *j=new PutJob(Clone(),get_args,cont);
   return j;
}

CMD(mget)
{
   Job *j=new mgetJob(Clone(),args);
   args=0;
   return j;
}

CMD(mput)
{
   Job *j=new mputJob(Clone(),args);
   args=0;
   return j;
}

CMD(shell)
{
   Job *j;
   if(args->count()==1)
      j=new SysCmdJob(0);
   else
      j=new SysCmdJob(args->getarg(1));
   return j;
}

CMD(rm)
{
   Job *j=new rmJob(Clone(),args);
   args=0;
   return j;
}
CMD(mkdir)
{
   Job *j=new mkdirJob(Clone(),args);
   args=0;
   return j;
}

CMD(source)
{
   if(args->count()<2)
   {
      printf("Usage: source <file>\n");
      return 0;
   }
   source=new FileStream(args->getarg(1),O_RDONLY);
   source_buf=source_buf_ptr=0;
   source_buf_size=0x1000/2;
   return 0;
}

CMD(jobs)
{
   int opt;
   args->rewind();
   int v=1;
   while((opt=args->getopt("+v"))!=EOF)
   {
      switch(opt)
      {
      case('v'):
	 v++;
	 break;
      case('?'):
	 printf("Usage: jobs [-v] [-v] ...\n");
	 return 0;
      }
   }
   Job::ListJobs(v);
   exit_code=0;
   return 0;
}

CMD(cd)
{
   if(args->count()!=2)
   {
      fprintf(stderr,"Usage: cd remote-dir\n");
      return 0;
   }

   char c;
   if(sscanf(args->getarg(1),"%*[a-z]://%*[^/]%c",&c)==1)
      return do_open();

   GetBetter();
   session->Chdir(args->getarg(1));
   builtin=BUILTIN_CD;
   return this;
}

CMD(pwd)
{
   const char *cwd=session->GetCwd();
   printf("%s\n",cwd[0]?cwd:"~");
   exit_code=0;
   return 0;
}

static void  move_to_background()
{
   fflush(stdout);
   fflush(stderr);
   switch(fork())
   {
   case(0): // child
   {
      signal(SIGINT,SIG_IGN);
      signal(SIGQUIT,SIG_IGN);
      signal(SIGHUP,SIG_IGN);
      signal(SIGTSTP,SIG_IGN);

      char *home=getenv("HOME");
      if(!home) home=".";
      char *log=(char*)alloca(strlen(home)+1+9+1);
      sprintf(log,"%s/.lftp_log",home);
      int fd=open(log,O_WRONLY|O_APPEND|O_CREAT,0700);
      if(fd>=0)
      {
	 dup2(fd,1);
	 dup2(fd,2);
	 if(fd!=1 && fd!=2)
	    close(fd);
      }
      for(;;)
      {
	 SMTask::Schedule();
	 if(Job::NumberOfJobs()==0)
	    break;
	 SMTask::Block();
      }
      exit(0);
   }
   default: // parent
      printf("Moving into background to complete transfers...\n");
      fflush(stdout);
      _exit(0);
   case(-1):
      perror("fork()");
   }
}

CMD(exit)
{
   if(args->count()>=2)
   {
      if(sscanf(args->getarg(1),"%i",&prev_exit_code)!=1)
      {
	 printf("Usage: %s [<exit_code>]\n",args->a0());
	 return 0;
      }
   }
   if(Job::NumberOfJobs()>0)
      move_to_background();
   exit(prev_exit_code);
}

CmdExec *CmdExec::debug_shell=0;
void CmdExec::debug_callback(char *msg)
{
   if(!debug_shell)
      return;
   if(debug_shell->status_line==0)
      return;
   debug_shell->status_line->WriteLine(msg);
   if(debug_shell->waiting && debug_shell->interactive)
      debug_shell->ShowRunStatus(debug_shell->status_line);
}

CMD(debug)
{
   FILE	 *new_dfile=0;//stderr;
   void	 (*new_cb)(char *)=&debug_callback;
   int	 new_dlevel=9;

   debug_shell=this;

   if(args->count()>1)
   {
      if(!strcmp(args->getarg(1),"off"))
      {
	 new_cb=0;
	 new_dfile=0;
	 new_dlevel=0;
      }
      else
      {
	 new_dlevel=atoi(args->getarg(1));
      }
   }
   if(new_cb)
   {
      printf("debug level %d\n",new_dlevel);
   }
   else
   {
      printf("debug off\n");
   }
   GetBetter();
   session->SetDebug(new_dfile,new_dlevel,new_cb);
   exit_code=0;
   return 0;
}

CMD(user)
{
   char	 *pass;
   if(args->count()<2 || args->count()>3)
   {
      fprintf(stderr,"Usage: %s userid [pass]\n",args->getarg(0));
      return 0;
   }
   if(args->count()==2)
      pass=GetPass("Password: ");
   else
      pass=args->getarg(2);
   if(pass)
      session->Login(args->getarg(1),pass);
   exit_code=0;
   return 0;
}
CMD(anon)
{
   session->AnonymousLogin();
   exit_code=0;
   return 0;
}

void unquote(char *buf,char *str)
{
   while(*str)
   {
      if(*str=='\"' || *str=='\\')
	 *buf++='\\';
      *buf++=*str++;
   }
   *buf=0;
}

CMD(open)
{
   bool	 debug=false;
   int	 port=-1;
   char	 *host=NULL;
   char  *path=NULL;
   char	 *user=NULL;
   char	 *pass=NULL;
   url_components uc;
   int	 c;
   NetRC::Entry *nrc=0;
   char	 *cmd_to_exec=0;

   args->rewind();
   while((c=args->getopt("u:p:e:d"))!=EOF)
   {
      switch(c)
      {
      case('p'):
	 if(!isdigit(optarg[0]))
	 {
	    struct servent *serv=getservbyname(optarg,"tcp");
	    if(serv==NULL)
	    {
	       fprintf(stderr,"%s: %s - no such tcp service\n",args->getarg(0),optarg);
	       return 0;
	    }
	 }
	 else
	    port=atoi(optarg);
	 break;
      case('u'):
         user=optarg;
         pass=strchr(optarg,',');
	 if(pass==NULL)
	    pass=strchr(optarg,' ');
	 if(pass==NULL)
   	    break;
	 *pass=0;
	 pass++;
         break;
      case('d'):
	 debug=true;
	 break;
      case('e'):
	 cmd_to_exec=optarg;
	 break;
      case('?'):
	 fprintf(stderr,"Usage: %s [-p port] [-u user[,pass]] host\n",
	    args->getarg(0));
	 return 0;
      }
   }

   if(optind<args->count())
      host=args->getarg(optind++);

   if(host)
   {
      if(parse_url(host,&uc))
      {
	 if(strcmp(uc.proto,"ftp"))
	 {
	    fprintf(stderr,"%s: %s - not supported protocol\n",args->getarg(0),uc.proto);
	    return 0;
	 }
	 if(uc.user[0] && !user)
	    user=uc.user;
	 host=uc.host;
	 if(uc.port!=-1 && port==-1)
	    port=uc.port;
	 if(uc.path[0] && !path)
	    path=uc.path;
      }
      nrc=NetRC::LookupHost(host);
      if(nrc)
      {
	 if(nrc->user && !user)
	 {
	    user=nrc->user;
	    if(nrc->pass)
	       pass=nrc->pass;
	 }
      }
      if(port==-1)
	 port=21;
   }
   if(user)
   {
      if(!pass)
	 pass=GetPass("Password: ");
      if(!pass)
	 fprintf(stderr,"%s: GetPass() failed: %s -- assume anonymous login\n",
	    args->getarg(0),strerror(errno));
      else
	 session->Login(user,pass);
   }
   if(host)
   {
      session->Connect(host,port);
      builtin=BUILTIN_OPEN;
      session->SetFlag(Ftp::SYNC_MODE,sync_mode);
      session->SetFlag(Ftp::NOREST_MODE,norest_mode);
   }
   if(nrc)
      delete nrc;

   if(cmd_to_exec)
      PrependCmd(cmd_to_exec);

   if(path)
   {
      char *s=(char*)alloca(strlen(path)*2+10);
      strcpy(s,"&& cd \"");
      unquote(s+strlen(s),path);
      strcat(s,"\"\n");
      PrependCmd(s);
   }

   if(debug)
      PrependCmd("debug\n");

   if(host)
      return this;
   exit_code=0;
   return 0;
}

CMD(kill)
{
   if(args->count()<2)
   {
      printf("Usage: %s <jobno> ... | all\n",args->getarg(0));
      return 0;
   }
   if(!strcmp(args->getarg(1),"all"))
   {
      Job::KillAll();
      exit_code=0;
      return 0;
   }
   args->rewind();
   for(;;)
   {
      char *arg=args->getnext();
      if(arg==0)
	 break;
      if(!isdigit(arg[0]))
      {
	 fprintf(stderr,"%s: `%s' - not a number\n",args->getarg(0),arg);
      	 continue;
      }
      int n=atoi(arg);
      if(Job::Running(n))
	 Job::Kill(n);
      else
	 fprintf(stderr,"%s: %d: No such job\n",args->getarg(0),n);
   }
   exit_code=0;
   return 0;
}

const char * const CmdExec::var_list[]=
{
   "ls-default","prompt","remote-completion","norest-mode","sync-mode",
   "long-running","timeout","redial-interval",
   NULL
};

CMD(set)
{
   if(args->count()<3)
   {
      if(args->count()!=1)
      {
	 printf("Usage: %s <variable> <value>\n",args->getarg(0));
	 return 0;
      }
      printf("set long-running %d\n",long_running);
      printf("set ls-default \"%s\"\n",var_ls);
      printf("set norest-mode %s\n",var_bool2str(norest_mode));
      printf("set prompt \"%s\"\n",var_prompt);
      printf("set redial-interval %d\n",var_redial_interval);
      printf("set remote-completion %s\n",var_bool2str(remote_completion));
      printf("set sync-mode %s\n",var_bool2str(sync_mode));
      printf("set timeout %d\n",var_timeout);
      exit_code=0;
      return 0;
   }
   exit_code=0;
   const char *v;
   if(!find_var_name(args->getarg(1),var_list,&v))
   {
      printf("Unknown variable: `%s'. Use `set' to look at all variables.\n",
	 args->getarg(1));
      exit_code=1;
      return 0;
   }
   if(!v)
   {
      printf("Ambiguous variable name: `%s'. Use `set' to look at all variables.\n",
	 args->getarg(1));
      exit_code=1;
      return 0;
   }

   char *c=args->Combine(2);

   if(!strcmp(v,"ls-default"))
   {
      free(var_ls);
      var_ls=c;
      return 0;
   }
   if(!strcmp(v,"prompt"))
   {
      free(var_prompt);
      var_prompt=c;
      return 0;
   }
   if(!strcmp(v,"remote-completion"))
   {
      remote_completion=var_str2bool(c);
      free(c);
      return 0;
   }
   if(!strcmp(v,"sync-mode"))
   {
      sync_mode=var_str2bool(c);
      session->SetFlag(Ftp::SYNC_MODE,sync_mode);
      free(c);
      return 0;
   }
   if(!strcmp(v,"norest-mode"))
   {
      norest_mode=var_str2bool(c);
      session->SetFlag(Ftp::NOREST_MODE,norest_mode);
      free(c);
      return 0;
   }
   if(!strcmp(v,"long-running"))
   {
      long_running=atoi(c);
      if(long_running<0)
	 long_running=0;
      free(c);
      return 0;
   }
   if(!strcmp(v,"timeout"))
   {
      var_timeout=atoi(c);
      if(var_timeout<30)
	 var_timeout=30; // set reasonable low limit
      session->SetTimeout(var_timeout);
      free(c);
      return 0;
   }
   if(!strcmp(v,"redial-interval"))
   {
      var_redial_interval=atoi(c);
      if(var_redial_interval<0)
	 var_redial_interval=0;
      session->SetSleep(var_redial_interval);
      free(c);
      return 0;
   }
   abort(); // we should have handled all cases
}

CMD(alias)
{
   if(args->count()<2)
   {
      Alias::List();
   }
   else if(args->count()==2)
   {
      Alias::Del(args->getarg(1));
   }
   else
   {
      char *val=args->Combine(2);
      Alias::Add(args->getarg(1),val);
      free(val);
   }
   exit_code=0;
   return 0;
}

CMD(wait)
{
   if(args->count()!=2)
   {
      printf("Usage: wait <jobno>\n");
      return 0;
   }
   args->rewind();
   char *jn=args->getnext();
   if(!isdigit(jn[0]))
   {
      fprintf(stderr,"wait: <jobno> must be a number\n");
      return 0;
   }
   Job *j=FindJob(atoi(jn));
   if(j==0)
   {
      fprintf(stderr,"wait: no such job %s\n",jn);
      return 0;
   }
   if(j->parent && j->parent->waiting==j)
   {
      fprintf(stderr,"wait: some other job waits for job %s\n",jn);
      return 0;
   }
   j->parent=0;
   return j;
}

CMD(site)
{
   if(args->count()<=1)
   {
      printf("Usage: site <site_cmd>\n");
      return 0;
   }
   Job *j=new SiteJob(Clone(),args->Combine(1),
      output?:new FDStream(1,"<stdout>"));
   output=0;
   return j;
}

CMD(subsh)
{
   CmdExec *e=new CmdExec(Clone());
   e->CopyEnv(this);

   char *c=args->getarg(1);
   e->FeedCmd(c);
   e->cmdline=(char*)xmalloc(strlen(c)+3);
   sprintf(e->cmdline,"(%s)",c);
   return e;
}

CMD(mirror)
{
   static struct option mirror_opts[]=
   {
      {"delete",no_argument,0,'e'},
      {"allow-suid",no_argument,0,'s'},
      {"include",required_argument,0,'i'},
      {"exclude",required_argument,0,'x'},
      {"time-prec",required_argument,0,'t'},
      {"only-newer",no_argument,0,'n'},
      {"no-recursion",no_argument,0,'r'},
      {"no-perms",no_argument,0,'p'},
      {0}
   };

   // mirror [--delete|-e] [--allow-suid|-s] remote local
   char cwd[1024];
   int opt;
   int	 flags=0;
   char *include=0;
   char *exclude=0;
   unsigned long prec=0;
   char	 ch;
   int n;

   args->rewind();
   while((opt=args->getopt_long("esi:x:t:nrp",mirror_opts,0))!=EOF)
   {
      switch(opt)
      {
      case('e'):
	 flags|=MirrorJob::DELETE;
	 break;
      case('s'):
	 flags|=MirrorJob::ALLOW_SUID;
	 break;
      case('r'):
	 flags|=MirrorJob::NO_RECURSION;
	 break;
      case('n'):
	 flags|=MirrorJob::ONLY_NEWER;
	 break;
      case('p'):
	 flags|=MirrorJob::NO_PERMS;
	 break;
      case('t'):
	 n=sscanf(optarg,"%lu%c",&prec,&ch);
	 if(n<1)
	 {
	 invalid_prec:
	    fprintf(stderr,
	       "mirror: --time-prec - invalid precision\n"
	       "mirror: Use `help mirror' to look at the syntax\n"); 
	    return 0;
	 }
	 if(n==1)
	    ch='s';
	 else if(ch=='m')
	    prec*=60;
	 else if(ch=='h')
	    prec*=60*60;
	 else if(ch=='d')
	    prec*=24*60*60;
	 else
	    goto invalid_prec;
	 break;
      case('x'):
	 exclude=optarg;
	 break;
      case('i'):
	 include=optarg;
	 break;
      case('?'):
	 return 0;
      }
   }

   args->back();
   const char *rcwd=args->getnext();
   if(!rcwd)
      rcwd=session->GetCwd();
   char *arg=args->getnext();
   if(arg)
   {
      strcpy(cwd,arg);
      create_directories(cwd);
   }
   else
   {
      if(getcwd(cwd,sizeof(cwd))==0)
      {
	 perror("getcwd()");
	 return 0;
      }
   }
   MirrorJob *j=new MirrorJob(Clone(),cwd,rcwd);
   j->SetFlags(flags,1);
   if(include)
   {
      if(j->SetInclude(include)==-1)
      {
	 delete j;
	 return 0;
      }
   }
   if(exclude)
   {
      if(j->SetExclude(exclude)==-1)
      {
	 delete j;
	 return 0;
      }
   }
   j->SetPrec((time_t)prec);
   return j;
}

CMD(mv)
{
   if(args->count()!=3)
   {
      printf("Usage: mv <file1> <file2>\n");
      return 0;
   }
   Job *j=new mvJob(Clone(),args->getarg(1),args->getarg(2));
   return j;
}

CMD(cache)  // cache control
{
   char *op=args->getnext();
   
   exit_code=0;
   if(!op || !strcmp(op,"stat"))
      LsCache::List();
   else if(!strcmp(op,"flush"))
      LsCache::Flush();
   else if(!strcmp(op,"on"))
      LsCache::On();
   else if(!strcmp(op,"off"))
      LsCache::Off();
   else if(!strcmp(op,"size"))
   {
      op=args->getnext();
      if(!op)
      {
	 fprintf(stderr,"%s: Operand missed for size\n",args->a0());
	 exit_code=1;
	 return 0;
      }
      long lim=-1;
      if(strcmp(op,"unlim") && sscanf(op,"%ld",&lim)!=1)
      {
	 fprintf(stderr,"%s: Invalid number for size\n",args->a0());
	 exit_code=1;
	 return 0;
      }
      LsCache::SetSizeLimit(lim);
   }
   else
   {
      printf("Invalid command. Try `help %s' for more information.\n",args->a0());
      exit_code=1;
   }
   return 0;
}
